/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2006 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @author Martin Lazarov */

#include "LRS_ExpressionParser.h"

#include <memory>
#include <sstream>


namespace {

using namespace LRS;

struct ExpressionParserImpl {
  enum { ttEof = -1, ttString = 0 };
  std::string _token, _used;
  int _tt;
  std::istringstream _in;
  
  Expression* parse(const std::string& expr) {
    _in.str(expr);

    if (nextToken() == ttEof) {
      throwError("unexpected EOF");
    }
    std::auto_ptr<Expression> root(parseExpr());
    if (_tt != ttEof) {
      throwError("expected EOF");
    }
    return root.release();
  }
  
  Expression* parseExpr() {
    std::auto_ptr<Expression> expr(new Expression());
    while (true) {
      switch (_tt) {
        case '^':
          nextToken();
          expr->addTerm(new ExCont(parseExpr()));
          break;
        case '{':
          nextToken();
          expr->addTerm(new InCont(parseExpr()));
          if (_tt != '}')
            throwError("expected '}'");
          nextToken();
          break;
        case '[':
          nextToken();
          expr->addTerm(new SqCont(parseExpr()));
          if (_tt != ']')
            throwError("expected ']'");
          nextToken();
          break;
        case ttString:
          expr->addTerm(parseId());
          break;
        default:
          throwError("unexpected token");
      }
      if (_tt != ',')
        break;
      nextToken();
    }
    return expr.release();
  }
  
  Term* parseId() {
    std::string t = _token;
    std::auto_ptr<Term> term;
    
    if (nextToken() == '(') {
      nextToken();
      std::auto_ptr<Expression> args(parseExpr());
      if (_tt != ')')
        throwError("expected ')'");
      ListOfTerms* subTerms = 0;
      if (nextToken() == ':') 
        subTerms = parseSubTerms();
      term.reset(new Functor(t, args.get(), subTerms));
      args.release();
    }
    else {
      char c = t[0];
      if (c >= 'a' && c <= 'z')
        term.reset(new Var(t));
      else {
        ListOfTerms* subTerms = 0;
        if (_tt == ':') 
          subTerms = parseSubTerms();
        term.reset(new MetaVar(t, subTerms));
      }
    }
    
    return term.release();
  }
  
  Expression* parseSubTerms() {
    if (nextToken() != '[')
      throwError("expected '['");
    nextToken();
    std::auto_ptr<Expression> subTerms(parseExpr());
    if (_tt != ']')
      throwError("expected ']'");
    nextToken();
    return subTerms.release();
  }

  void throwError(const std::string& msg) {
    throw ExpressionParser::SyntaxError(msg, _used);
  }
  
  static bool isIdStart(char t) {
    return t == '_' || t >= 'A' && t <= 'Z' || t >= 'a' && t <= 'z';
  }
  
  static bool isIdPart(char t) {
    return isIdStart(t) || t >= '0' && t <= '9';
  }
  
  void skipWhiteSpace() {
    while (isspace(_in.peek())) {
      _in.get();
    }
  }
  
  int nextToken() {
    skipWhiteSpace();
    int t = _in.get();
    if (t == -1) {
      _tt = ttEof;
      _token.clear();
    }
    else if (isIdStart(t)) {
      _tt = ttString;
      // read the whole string
      _token = (char) t;
      while (isIdPart(t = _in.get())) {
        _token += (char) t;
      }
      _in.unget();
    }
    else {
      _tt = t;
      _token = (char) t;
    }
    _used += _token;
    return _tt;
  }
};

}

namespace LRS {
  /**
    Parses a CLLRS string.
      
    The input string has to conform to the following grammar:
    
    LRS-Expression:
      ListOfTerms .

    ListOfTerms:
      Term | ListOfTerms "," Term .
    
    Term:
      Var | MetaVar | Functor | ExCont | InCont | SqCont .
    
    Var:
      / [a-z][A-Za-z0-9_]* / .

    MetaVar:
      / [A-Z_][A-Za-z0-9_]* / SubTerms? .
    
    Functor:
      / [A-Za-z_][A-Za-z0-9_]* / "(" ListOfTerms ")" SubTerms? .
    
    SubTerms:
      ":[" ListOfTerms "]" .

    ExCont:
      "^" ListOfTerms .

    InCont:
      "{" ListOfTerms "}" .
    
    SqCont:
      "[" ListOfTerms "]" .

  */
  Expression* ExpressionParser::parse(const std::string& expr) {
    return ExpressionParserImpl().parse(expr);
  }
  
  ExpressionParser::SyntaxError::SyntaxError(
    const std::string& msg, const std::string& used
    ) : GeneralError("Syntax error: ") {
    _msg += msg;
    _msg += "\n";
    _msg += "-- after: ";
    _msg += used;
  }
}
