/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2006 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @author Martin Lazarov */

#include "LRS_MainWidget.h"
#include "LRS_NodeWidget.h"

#include <qpainter.h>
#include <qpaintengine.h>
#include <qprinter.h>
#include <qprintdialog.h>
#include <qfont.h>
#include <qglobal.h>
#include <qmessagebox.h>
#include <qvariant.h>

namespace {
  const int MIN_FONT_SIZE = 2;
  const int MAX_FONT_SIZE = 72;
  
  const int MIN_NODE_DISTANCE = 0;
  const int MAX_NODE_DISTANCE = 600;
  
  const int FONT_SIZE_INC = 1;
  const int NODE_DISTANCE_INC = 5;
}

namespace LRS {

MainWidget::MainWidget(
  const std::string& lrsExpr,
  const std::string& fontName, int fontSize,
  int minHDistance, int minVDistance, int borderSize
  )
  : _layout(borderSize, borderSize, minHDistance, minVDistance),
    _borderSize(borderSize)
{
  QFont f(fontName.c_str(), fontSize);
  //f.setItalic(true);
  //QMessageBox::critical(0, "font", QVariant(f).toString());
  //setFont(QVariant(QVariant(f).toString()).value<QFont>());
  setFont(f);
  _root = new NodeWidget(lrsExpr, this);
  updateLayout();
}

void MainWidget::updateLayout() {
  _layout.layoutNode(_root);
  const int maxX = _root->treeWidth()  + _borderSize * 2;
  const int maxY = _root->treeHeight() + _borderSize * 2;
  setMaximumSize(maxX, maxY);
  resize(maxX, maxY);
  update();
}

void MainWidget::incFontSize() {
  if (font().pointSize() + FONT_SIZE_INC > MAX_FONT_SIZE)
    return;
  QFont newFont(font());
  newFont.setPointSize(newFont.pointSize() + FONT_SIZE_INC);
  setFont(newFont);
  updateLayout();
}

void MainWidget::decFontSize() {
  if (font().pointSize() - FONT_SIZE_INC < MIN_FONT_SIZE)
    return;
  QFont newFont(font());
  newFont.setPointSize(font().pointSize() - FONT_SIZE_INC);
  setFont(newFont);
  updateLayout();
}

void MainWidget::incHDistance() {
  if (_layout.hDistance() + NODE_DISTANCE_INC > MAX_NODE_DISTANCE)
    return;
  _layout.setHDistance(_layout.hDistance() + NODE_DISTANCE_INC);
  updateLayout();
}

void MainWidget::decHDistance() {
  if (_layout.hDistance() - NODE_DISTANCE_INC < MIN_NODE_DISTANCE)
    return;
  _layout.setHDistance(_layout.hDistance() - NODE_DISTANCE_INC);
  updateLayout();
}

void MainWidget::incVDistance() {
  if (_layout.vDistance() + NODE_DISTANCE_INC > MAX_NODE_DISTANCE)
    return;
  _layout.setVDistance(_layout.vDistance() + NODE_DISTANCE_INC);
  updateLayout();
}

void MainWidget::decVDistance() {
  if (_layout.vDistance() - NODE_DISTANCE_INC < MIN_NODE_DISTANCE)
    return;
  _layout.setVDistance(_layout.vDistance() - NODE_DISTANCE_INC);
  updateLayout();
}

void MainWidget::paintOnDevice(QPaintDevice* dev) {
  QPainter painter(dev);
  painter.setRenderHint(QPainter::Antialiasing);
  drawEdges(painter, _root);
}
  
void MainWidget::paintEvent(QPaintEvent*) {
  paintOnDevice(this);
}

void MainWidget::drawEdges(QPainter& painter, NodeWidget* w) {
  //painter.drawRect(w->x(), w->y(), w->width(), w->height());
  int x  = w->x() + w->section(0)->x();
  int y1 = w->y() + w->height();
  for (unsigned i = 0; i < w->numSections(); ++i) {
    const NodeWidget::Section* s = w->section(i);
    for (unsigned j = 0; j < s->children().size(); ++j) {
      NodeWidget* v = s->children()[j];
      if (v->isVisible()) {
        int x1 = x + s->width() / 2;
        int x2 = v->x() + v->width() / 2;
        int y2 = v->y();
        painter.drawLine(x1, y1, x2, y2);
        drawEdges(painter, v);
      }
    }
    if (s->hasCollapsedChild()) {
      // not fully expanded sections are maked
      // with a red dot below them
      //
      painter.fillRect(x + s->width() / 2 - 3, y1 - 3, 6, 6, Qt::red);
    }
    x += s->width();
  }
}

void MainWidget::print(QPrinter* p) {
  QRect r(p->pageRect());
  if (r.width() >= width() && r.height() >= height()) {
    r.setSize(size());
  }
  else {
    // scale down
    if (r.width() < width()) {
      double f = double(r.width()) / width();
      r.setWidth(width());
      r.setHeight(int(f * height()));
    }
    if (r.height() < height()) {
      double f = double(r.height()) / height();
      r.setHeight(height());
      r.setWidth(int(f * width()));
    }
  }
  QPainter(p).drawPixmap(r, QPixmap::grabWidget(this));
}

void MainWidget::print() {
  QPrinter p;
  QPrintDialog pd(&p, this);
  if (pd.exec() == QDialog::Accepted) {
    print(&p);
  }
}

}
