/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2006 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @author Martin Lazarov */

#include "LRS_NodeWidget.h"
#include "LRS_MainWidget.h"

#include "LRS_ExpressionParser.h"

#include <algorithm>
#include <functional>
#include <memory>

#include <qpainter.h>
#include <qevent.h>

//#include <LRS_Debug.h>
//#include <sstream>
//#include <iostream>

namespace LRS {

using namespace std;

struct NodeWidgetCreatingVisitor : TermVisitor, NonCopyable {
  string _label;
  NodeWidget* _w;
  
  NodeWidgetCreatingVisitor(Expression* expr, NodeWidget* w)
    : _w(w) {
    processListOfTerms(expr);
  }
  NodeWidgetCreatingVisitor(Term* t, NodeWidget* w)
    : _w(w) {
    t->accept(this);
  }
  virtual ~NodeWidgetCreatingVisitor() {
    addStaticSection();
    _w->updateLayout();
  }
  void addStaticSection() {
    if (_label.empty()) return;
    NodeWidget::Section* s = new NodeWidget::Section(_label.c_str(), true, _w);
    _label.erase();
    _w->_sections.push_back(s);
  }
  void addDynamicSection(const string& name) {
    addStaticSection();
    NodeWidget::Section* s = new NodeWidget::Section(name.c_str(), false, _w);
    _w->_sections.push_back(s);
  }
  void processSubTerms(ComplexTerm* t) {
    NodeWidget::Section* s = _w->_sections.back();
    // create from each sub-term a new node-widget
    // and append it as a child of the current section
    for (unsigned i = 0; i < t->subTerms()->size(); ++i) {
      Term* subTerm = t->subTerms()->at(i);
      NodeWidget* w = new NodeWidget(s, _w);
      NodeWidgetCreatingVisitor(subTerm, w);
      s->addChild(w);
      _w->_children.push_back(w);
    }
  }
  void processListOfTerms(const ListOfTerms* ls) {
    for (unsigned i = 0; i < ls->size(); ++i) {
      if (i) _label += ",";
      ls->at(i)->accept(this);
    }
  }
  void visit(Var* t) {
    _label += t->name();
  }
  void visit(Functor* t) {
    TermVisitor::visit(t);
    _label += "(";
    processListOfTerms(t->arguments());
    _label += ")";    
  }
  void visit(ComplexTerm* t) {
    if (!t->subTerms())
      _label += t->name();
    else {
      addDynamicSection(t->name());
      processSubTerms(t);
    }
  }
  void visit(ExCont* t) {
    _label += "^";
    processListOfTerms(t->innerTerms());
  }
  void visit(InCont* t) {
    _label += "{";
    processListOfTerms(t->innerTerms());
    _label += "}";
  }
  void visit(SqCont* t) {
    _label += "[";
    processListOfTerms(t->innerTerms());
    _label += "]";
  }
};

void NodeWidget::Section::expandChildren() {
  for_each(
    _children.begin(), _children.end(), 
    mem_fun(&NodeWidget::expand)
    );
  clearHasCollapsedChild();
  static_cast<NodeWidget*>(parentWidget())->childrenSync();
}

void NodeWidget::Section::showChildren() {
  for_each(
    _children.begin(), _children.end(), 
    mem_fun(&NodeWidget::_show)
    );
}

void NodeWidget::Section::hideChildren() {
  for_each(
    _children.begin(), _children.end(), 
    mem_fun(&NodeWidget::_hide)
    );
}

void NodeWidget::Section::mouseReleaseEvent(QMouseEvent* me) {
  if (me->button() == Qt::RightButton) {
    static_cast<NodeWidget*>(parentWidget())->collapse();
  }
  else if (me->button() == Qt::LeftButton) {
    if (hasCollapsedChild()) {
      expandChildren();
    }
  }
}
  
NodeWidget::NodeWidget(const string& lrsExpr, QWidget* parent)
  : QWidget(parent), _parentNode(0), _parentSection(0), _collapsed(false),
    _treeW(0), _treeH(0)
{
  auto_ptr<Expression> expr(ExpressionParser::parse(lrsExpr));
  NodeWidgetCreatingVisitor(expr.get(), this);
}

NodeWidget::NodeWidget(Section* parentSection, NodeWidget* parentNodeWidget)
  : QWidget(parentNodeWidget->parentWidget()), 
    _parentNode(parentNodeWidget), _parentSection(parentSection),
    _collapsed(false), _treeW(0), _treeH(0)
{}

void NodeWidget::collapse() {
  if (!_parentSection) return;
  _hide();
  _collapsed = true;
  _parentSection->setHasCollapsedChild();
  _parentNode->childrenSync();
}

void NodeWidget::expand() {
  if (!_collapsed) return;
  _collapsed = false;
  _show();
}

void NodeWidget::_hide() {
  if (_collapsed) return;
  hide();
  for_each(
    _sections.begin(), _sections.end(),
    mem_fun(&Section::hideChildren)
    );
}

void NodeWidget::_show() {
  if (_collapsed) return;
  show();
  for_each(
    _sections.begin(), _sections.end(),
    mem_fun(&Section::showChildren)
    );
}

void NodeWidget::childrenSync() {
  _children.clear();
  vector<Section*>::iterator iter = _sections.begin(),
    iter_end = _sections.end();
  for (; iter != iter_end; ++iter) {
    const Section* s = *iter;
    for (unsigned i = 0; i < s->children().size(); ++i) {
      NodeWidget* w = s->children()[i];
      if (!w->_collapsed)
        _children.push_back(w);
    }
  }
  static_cast<LRS::MainWidget*>(parentWidget())->updateLayout();
}

void NodeWidget::updateLayout() {
  const int WIDGET_BORDER_WIDTH = 5;
  
  int h = 0;
  int w = WIDGET_BORDER_WIDTH;
  
  for (unsigned i = 0; i < _sections.size(); ++i) {
    Section* s = _sections[i];
    s->adjustSize();
    s->move(w, WIDGET_BORDER_WIDTH);
    w += s->width();
    if (h < s->height())
      h = s->height();
  }
  w += WIDGET_BORDER_WIDTH;
  h += WIDGET_BORDER_WIDTH * 2;
  resize(w, h);
}

void NodeWidget::paintEvent(QPaintEvent*) {
  //QPainter painter(this);
  //painter.setPen(Qt::DotLine);
  //painter.drawRect(0, 0, width() - 1, height() - 1);
}

#if QT_VERSION >= 0x40000

void NodeWidget::changeEvent(QEvent* ev) {
  if (ev->type() == QEvent::FontChange)
    updateLayout();
}

#else

void NodeWidget::fontChange(const QFont&) {
  updateLayout();
}

#endif

}
