/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2006 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @author Martin Lazarov */

#ifndef LRS_UTIL_H
#define LRS_UTIL_H

#include <vector>
#include <string>
#include <stdexcept>

namespace LRS {

/** Prevents instances of derived classes from being passed by value.
    @see http://www.boost.org/libs/utility/utility.htm#Class_noncopyable
*/
struct NonCopyable {
  NonCopyable() {}
private:
  NonCopyable(const NonCopyable&);
  NonCopyable& operator=(const NonCopyable&);
};

/** An extension of std::vector in the spirit of std::auto_ptr. */
template <class T>
struct AutoPtrVector : std::vector<T*>, NonCopyable {
  virtual ~AutoPtrVector() {
    typedef std::vector<T*> base;
    for (typename base::size_type i = 0; i < base::size(); ++i)
      delete base::at(i);
  }
};

/** A base class for exceptions thrown by the lrsviewer code. */
struct GeneralError : std::exception { 
  GeneralError(const char* err, const char* desc = 0) : _msg(err) {
    if (desc) {
      _msg += "\n";
      _msg += desc;
    }
  }
  ~GeneralError() throw()
    {}
  const char* what() const throw()
    { return _msg.c_str(); }
protected:  
  std::string _msg;
};

/** Trims the whitespace from the beginning of a string. */
inline
std::string& ltrim(std::string& s) {
  std::string::size_type i = 0;
  for (; i < s.length() && isspace(s[i]); ++i);
  if (i == s.size())
    s.erase();
  else {
    std::string::size_type j = 0;
    for (; i < s.length(); ++i, ++j)
      s[j] = s[i];
    s.erase(j);
  }
  return s;
}

/** Trims the whitespace from the end of a string. */
inline
std::string& rtrim(std::string& s) {
  std::string::size_type i = s.length();
  for (; i > 0 && isspace(s[i - 1]); --i);
  s.erase(i);
  return s;
}

/** Trims the whitespace from beginning and the end of a string. */
inline
std::string& trim(std::string& s) {
  return rtrim(ltrim(s));
}
  
} // namespace LRS

#endif // !LRS_UTIL_H
