/*  $Id$

    This file is part of the

    GRALE: a grammar visualization tool

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2005 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/** @file

    This file contains application-wide available utiliy functions.

    @author Holger Wunsch */


#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include <q3filedialog.h>
#include <qmessagebox.h>

#include <qfile.h>
#include <qdir.h>
#include <qapplication.h>
#include <qfiledialog.h>

#include "tools.h"

/** Presents a "Save As" dialog to the user.
    @param parent the parent widget.
    @param caption the dialog's caption.
    @param existCheck if <code>true</code>, the dialog will show a warning dialog
      if a file with the selected name already exists.
    @param filter the filter string (see <code>QFileDialog</code>).
    @return the filename the user selected, or <code>QString::null</code> if
      the user cancelled the "Save As" dialog or the warning dialog about an
      existing file. */
QString getSaveFilename(QWidget* parent, const QString& caption, bool existCheck, 
                        const QString& filter) {
  QString fn;
  QFileDialog::Option options =
    existCheck ? QFileDialog::Option(0) : QFileDialog::DontConfirmOverwrite;

  fn = QFileDialog::getSaveFileName(parent, caption, ".", filter, 0, options);
  if (fn.isNull())
    return QString::null;
  
  QDir::setCurrent(getDirPath(fn));
  return fn;
}

QString getDirPath(const QString& filepath) {
  QString d(filepath);
  int i = d.lastIndexOf("/");
  if (i == -1) {
    i = d.lastIndexOf("\\");
    if (i == -1)
      return ".";
  }
  d.chop(d.length() - i);
  return d;
}

/**
 * This function returns a temporary file name based on
 * the current process ID, the current time and a
 * random number.
 * @return the temporary name, a pointer to a static buffer.
 */
const char* getTempName(void) {
  static char buffer[500];

  pid_t pid = getpid();
  time_t tm = time(NULL);

  srand(pid + tm);
  int rnd = rand() ^ pid;
  const char* tmpDir = getenv("TEMP");
  if (tmpDir == 0)
    tmpDir = "/tmp";

  sprintf(buffer, "%s/GRALE%u", tmpDir, rnd);

  return buffer;
}

namespace GRALE {

_Q3ToolBar::_Q3ToolBar(Q3MainWindow* parent) : Q3ToolBar(parent) {}
  
QToolButton* _Q3ToolBar::insertButton(const QPixmap& pixmap, int id, const char *signal,
    const QObject *receiver, const char *slot,
    bool enabled, const QString& text)
{
  QToolButton* b = new QToolButton(this);
  b->setIcon(pixmap);
  b->setEnabled(enabled);
  b->setToolTip(text);
  QObject::connect(b, signal, receiver, slot);
  _buttons[id] = b;
  return b;
}
  
void _Q3ToolBar::setItemEnabled(int itemId, bool enabled) {
  QMap<int,QToolButton*>::const_iterator it = _buttons.find(itemId);
  if (it != _buttons.end()) {
    it.value()->setEnabled(enabled);
  }
  else {
    error_message_box(parentWidget(), "toolbar error", "unknown toolbar item-id: %d", itemId);
  }
}
  
void _Q3ToolBar::insertSeparator(int, int) {
  addSeparator();
}

void _Q3ToolBar::insertSeparator() {
  addSeparator();
}

_IconLoader::_IconLoader() {
  _iconPath = qApp->applicationDirPath() + "/pics/";
}

QPixmap _IconLoader::loadIcon(QString iconName, int) {
  QString pngIcon(_iconPath + iconName + ".png");
  if (QFile::exists(pngIcon))
    return QPixmap(pngIcon);
  else
    return QPixmap(_iconPath + iconName + ".xpm");
}

inline
bool isInRange(char n, char lower, char upper) {
  return n >= lower && n <= upper;
}

bool isUtf8String(const char* s, int len) {
  if (len < 0)
    len = qstrlen(s);
  
  bool hasHighByte = false;
  
  for (int i = 0; i < len; ++i) {
    if (s[i] >= 0)
      continue;
      
    hasHighByte = true;
    
    int numEscapedBytes;
    
    if (isInRange(s[i], 0xc0, 0xdf))
      numEscapedBytes = 1;
    else if (isInRange(s[i], 0xb0, 0xef))
      numEscapedBytes = 2;
    else if (isInRange(s[i], 0xf0, 0xf4))
      numEscapedBytes = 3;
    else
      return false;
    
    int j = i;
    i += numEscapedBytes;
    if (i >= len)
      return false;

    while (j < i) {
      if (!isInRange(s[++j], 0x80, 0xbf))
        return false;
    }
  }
  
  return hasHighByte;
}

QString autoString(const char* s, int len) {
  if (isUtf8String(s, len))
    return QString::fromUtf8(s, len);
  
  return QString::fromLocal8Bit(s, len);
}

} // end of namespace GRALE
