/*  This package is a collection of utilities for performing operations
    on (mathematical) graphs.

    The P-representation of a graph is a list of (from-to) vertex
    pairs, where the pairs can be in any old order.  This form is
    convenient for input/output.

    The S-representation of a graph is a list of (vertex-neighbours)
    pairs, where the pairs are in standard order (as produced by
    keysort) and the neighbours of each vertex are also in standard
    order (as produced by sort).  This form is convenient for many
    calculations.

    p_to_s_graph(Pform, Sform) converts a P- to an S- representation.
    s_to_p_graph(Sform, Pform) converts an S- to a P- representation.

    warshall(Graph, Closure) takes the transitive closure of a graph
    in S-form.  (NB: this is not the reflexive transitive closure).

    p_transpose transposes a graph in P-form, cost O(|E|).
    s_transpose transposes a graph in S-form, cost O(|V|^2).

*/

:- use_module(library(ordsets),[
        ord_union/3
   ]),
   use_module(library(lists), [
	member/2,
	memberchk/2
   ]).

:- mode
	p_vertices(+, -),
	p_to_s_graph(+, -),
	    p_to_s_vertices(+, -),
	    p_to_s_group(+, +, -),
		p_to_s_group(+, +, -, -),
	s_to_p_graph(+, -),
	    s_to_p_graph(+, +, -, -),
	p_transpose(+, -),
	s_transpose(+, -),
	    s_transpose(+, -, ?, -),
		transpose_s(+, +, +, -),
	warshall(+, -),
	    warshall(+, +, -),
		warshall(+, +, +, -).

/*  A common pattern in this file involves two lists of vertex names:
    Neibs is the list of neighbours of some node,
    Vertices is the complete list of all vertex names for a graph.

	p(Neibs, Vertices, ...)

    p([], _, ...).
    p([V|Ns], [V|Vs], ...) :- !,
	{case when V is an element of Neibs}
	p(Ns, Vs, ...).
    p(Ns, [V|Vs], ...) :-
	{case when V is not an element of Neibs}
	p(Ns, Vs, ...).

    This works because Neibs and Vertices are in standard order, with Neibs
    being a subset of Vertices.  So the last clause corresponds to the case
    Ns = [N|_], N @> V.
*/

%   p_vertices(+P_Graph, ?Vertices)
%   unifies Vertices with the set of non-isolated vertices of P_Graph
%   represented as a list in standard order.  It is a characteristic
%   of P-representation that a vertex only appears if it is attached
%   to some edge.

p_vertices(P_Graph, Vertices) :-
	p_vertices_1(P_Graph, Bag),
	sort(Bag, Vertices).

p_vertices_1([], []).
p_vertices_1([From-To|Edges], [From,To|Vertices]) :-
	p_vertices_1(Edges, Vertices).


%   p_to_s_graph(+P_Graph, ?S_Graph)
%   converts a graph from P-representation to S-representation.
%   If the graph has N nodes and E edges, the cost is O(ElgE).

p_to_s_graph(P_Graph, S_Graph) :-
	sort(P_Graph, EdgeSet),
	p_vertices(EdgeSet, VertexSet),
	p_to_s_group(VertexSet, EdgeSet, S_Graph).


p_to_s_group([], _, []).
p_to_s_group([Vertex|Vertices], EdgeSet, G0) :-
	p_to_s_group(EdgeSet, Vertex, G0, G1, RestEdges),
	p_to_s_group(Vertices, RestEdges, G1).

p_to_s_group([V-X|Edges], V, [V-[X|Neibs]|G], G, RestEdges) :- !,
	p_to_s_group(Edges, V, Neibs, RestEdges).
p_to_s_group(Edges, V, [V-[]|G], G, Edges).

p_to_s_group([V-X|Edges], V, [X|Neibs], RestEdges) :- !,
	p_to_s_group(Edges, V, Neibs, RestEdges).
p_to_s_group(Edges, _, [], Edges).


%   s_to_p_graph(+S_Graph, ?P_Graph)
%   converts a graph from S-representation to P-representation.
%   A unique choice is made: edges appear in standard order.
%   If the graph has N nodes and E edges, the cost is O(E).

s_to_p_graph([], []).
s_to_p_graph([Vertex-Neibs|G], P_Graph) :-
	s_to_p_graph(Neibs, Vertex, P_Graph, Rest_P_Graph),
	s_to_p_graph(G, Rest_P_Graph).


s_to_p_graph([], _, P_Graph, P_Graph).
s_to_p_graph([Neib|Neibs], Vertex, [Vertex-Neib|P], Rest_P) :-
	s_to_p_graph(Neibs, Vertex, P, Rest_P).


%   warshall(+Graph, ?Closure) computes Closure from graph in
%   O(N**3) time.  To obtain the transitive closure of a
%   graph in P-representation, convert it to S-representation.
%   The closure will be the dominant cost.  Note that there
%   can be a lot of graphs with a given closure, so it wouldn't
%   make much sense to call this with a variable as first
%   argument.

warshall(Graph, Closure) :-
	warshall(Graph, Graph, Closure).


warshall([], Closure, Closure).
warshall([V-_|G], E, Closure) :-
	memberchk(V-Y, E),	%  Y := E(v)
	warshall(E, V, Y, NewE),
	warshall(G, NewE, Closure).


warshall([], _, _, []).
warshall([X-Neibs|G], V, Y, [X-NewNeibs|NewG]) :-
	memberchk(V, Neibs),
	!,
	ord_union(Neibs, Y, NewNeibs),
	warshall(G, V, Y, NewG).
warshall([X-Neibs|G], V, Y, [X-Neibs|NewG]) :-
	warshall(G, V, Y, NewG).


%   p_transpose(?P_Graph, ?P_Trans)
%   computes the transpose of a graph in P-representation.
%   It can solve for either argument.  The cost is O(E).

p_transpose([], []).
p_transpose([From-To|Edges], [To-From|Transpose]) :-
	p_transpose(Edges, Transpose).


%   s_transpose(+S_Graph, ?S_Trans)
%   computes the transpose of a graph in S-representation.
%   It can only be used one way around.  The cost is O(N**2).

s_transpose(S_Graph, Transpose) :-
	s_transpose(S_Graph, Base, Base, Transpose).

s_transpose([], [], Base, Base).
s_transpose([Vertex-Neibs|Graph], [Vertex-[]|RestBase], Base, Transpose) :-
	s_transpose(Graph, RestBase, Base, SoFar),
	transpose_s(SoFar, Neibs, Vertex, Transpose).

transpose_s([], [], _, []).
transpose_s([Neib-Trans|SoFar], [Neib|Neibs], Vertex,
		[Neib-[Vertex|Trans]|Transpose]) :- !,
	transpose_s(SoFar, Neibs, Vertex, Transpose).
transpose_s([Head|SoFar], Neibs, Vertex, [Head|Transpose]) :-
	transpose_s(SoFar, Neibs, Vertex, Transpose).

